#include "semaphore.h"
#include <queue>
#include "countingSemaphore.h"
#include "win32Semaphore.h"
#include "mutexLocker.h"
#include "SemaphoreIDGenerator.h"
#include "TDThread.h"
#include "deadlockWatchdog.h"
#include "threadData.h"
#include <typeinfo>
#include <string.h>
#include <assert.h>
#include <iostream>

countingSemaphore::countingSemaphore(int initialPermits, bool traceAndReplay)
 : semaphore(initialPermits) {
	if (traceAndReplay) {
	 char* env = getenv("MODE");
	 if (env!=NULL) {
		if (!(strcmp(env,"TRACE")))
			mode = TRACE;
		else if (!(strcmp(env,"REPLAY")))
			mode = REPLAY;
		else if (!(strcmp(env,"TEST")))
			mode = TEST;
		else if (!(strcmp(env,"SPECTEST")))
			mode = SPECTEST;
		else if (!(strcmp(env,"NONE")))
			mode = NONE;
		else mode = NONE;
	 }
	 else mode = NONE;
 	 if (mode!=TRACE && mode!=REPLAY  && mode != NONE) {
		std::cout << "Error: Only modes TRACE, REPLAY, and NONE are supported by semaphores." << std::endl;
		exit(1);
	 }
	 env = getenv("CONTROLLERS");
	 if (env!=NULL) {
		if (!(strcmp(env,"SINGLE")))
			numControllers = 1;
		else numControllers = 0;
	 }
	 else numControllers = 1;
	 env = getenv("RANDOMDELAY");
	 if (env!=NULL) {	
		 if	(!(strcmp(env,"ON")))
			 randomDelay = true;
		 else
			 randomDelay = false;
	 }
	 else randomDelay = false;

	 env = getenv("DEADLOCKDETECTION");
	 if (env!=NULL) {
		if (!(strcmp(env,"ON")))
			detectDeadlock = true;
		else
			detectDeadlock = false;
	 }
	 else detectDeadlock = false;

	 if (mode) {
		semaphoreName = generateID();
		std::string traceFileName;
		if (numControllers == 0) 
			traceFileName = semaphoreName;
		else
			traceFileName = "semaphoreandlock";
		if (mode == TRACE || mode == REPLAY)
			if (numControllers == 1) 
				control = instance(mode,numControllers,traceFileName);
			else
				control = new semaphoreAndLockTracingAndReplay(mode,numControllers,
							traceFileName);
	 }
	}
	else {mode = NONE; randomDelay = false;}

}

countingSemaphore::~countingSemaphore() {
	if (mode == TRACE || mode == REPLAY)
		if (numControllers = 0) 
		  delete control;
	// else control is a static shared by many threads - need to ensure it is 
	// deleted only once if you decide to delete it (need to lock control)

}		

std::string countingSemaphore::generateID() {
		const std::string thisClassName = "countingSemaphore";
		std::string semaphoreName;
		data d = threadData::instance().getThreadData(GetCurrentThreadId());	
		if (d.threadID != -1) 
		  semaphoreName = d.threadName;
		else
		  semaphoreName = "";
		int position = semaphoreName.find("_"); 
		if (position > -1 && (size_t) position < semaphoreName.length())
		  semaphoreName = semaphoreName + "_" + thisClassName;
		else
		  semaphoreName = "main_" + thisClassName;		
		std::string myID = SemaphoreIDGenerator::instance().getID(semaphoreName);
		return myID;
}

void countingSemaphore::P() {
	// each thread blocks on its own conditionVar
	if (mode == REPLAY) {
     	control->requestPermit((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
	}
    else if (randomDelay && mode == TRACE) {
	   Sleep(rand()%delayAmount);	// default delayAmount is 750
    }
	lock();
	--permits;
	if (permits>=0) {
		if (mode == REPLAY)
			control->releasePermit();
		else if (mode == TRACE) {
    		control->trace((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
		}
	   unlock();
	   return;
	}
    waitRecord rec;
    rec.s = pool.acquire();
    rec.ID = (threadData::instance().getThreadData(GetCurrentThreadId())).threadID;
    if (detectDeadlock && (mode == RT || mode == TRACE)) {
		TDThread::incNumBlockedThreads();
		deadlockWatchdog::changeStatus(GetCurrentThreadId(),"blocked on P operation of "+semaphoreName);
    }
    waitingP.push(rec); // otherwise append blocked thread
	unlock();
	rec.s->P();
    if (detectDeadlock && (mode == RT || mode == TRACE)) {
		deadlockWatchdog::changeStatus(GetCurrentThreadId(),"running");
    }
	// never block during replay
	if (mode == REPLAY)
		assert(false);
	pool.release(rec.s);
} // end P()

void countingSemaphore::V() {
	if (mode == REPLAY)
     	control->requestPermit((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
    else if (randomDelay && mode == TRACE) {
	   Sleep(rand()%delayAmount);	// default delayAmount is 750
    }
	lock(); // lock semaphore
	++permits;
	if (permits>0) {
		if (mode == REPLAY)
			control->releasePermit();
		else if (mode == TRACE) {
    		control->trace((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
		}
		unlock();
		return;
	}
    win32Semaphore* oldestConditionVar = (waitingP.front()).s;
	int oldestID = (waitingP.front()).ID;
	waitingP.pop();
	// never block during replay
	if (mode == REPLAY)
		assert(false);
	if (mode == TRACE) {
    	control->trace((threadData::instance().getThreadData(GetCurrentThreadId())).threadID);
	    control->trace(oldestID); // ID of thread that did P
	}
	oldestConditionVar->V();
    if (detectDeadlock && (mode == RT || mode == TRACE)) {
		TDThread::decNumBlockedThreads();
	}
    unlock(); // end synchronized (this) to avoid conditionVar.wait() while holding lock on this
} // end V()

void countingSemaphore::V_updateTimestamp() {}

bool countingSemaphore::doP() {
// Called by VP() operation; checks permit and unblocks a thread waiting
// in V() if necessary. Returns true if P() should block; false otherwise.
   --permits;
	if (permits>=0)
		return false;	// P does not block
	else
		return true;	// P() does block
}

void countingSemaphore::VP(semaphore* vSem) {
// execute {vSem.V(); this.P();} without any intervening 
// P() or V() operations on this or vSem.	

// lock semaphores in ascending order of IDs to prevent circular deadlock (i.e. T1 holds
// this's lock and waits for vSem's lock while T2 holds vSem's lock and waits for this's lock.)
   semaphore* first = this;
   semaphore* second = vSem;
   if (this->getSemaphoreID() > vSem->getSemaphoreID()) { 
     first = vSem;
     second = this;
   }
   first->lock(); 
   second->lock();
   //vSem.V() must not block
   //if (!strcmp(typeid(*vSem).name(),"binarySemaphore") && vSem->permits==1)
	if (vSem->permits==1 && vSem->isBinarySemaphore())
	   std::cout << "illegal VP operation" << std::endl;
   // perform vSem.V()
   vSem->V(); // it's okay that we already hold vSem's lock
   // perform this.P()
   bool blockingP = doP();
   if (!blockingP) {
     second->unlock();
     first->unlock();
     return;
   }
   waitRecord rec;
   rec.s = pool.acquire();
   rec.ID = (threadData::instance().getThreadData(GetCurrentThreadId())).threadID;
   waitingP.push(rec); // otherwise append blocked thread
   second->unlock();
   first->unlock();
   rec.s->P();
   pool.release(rec.s);
} // end VP()
