#pragma warning(disable:4786) 
#include <iostream>
#include "sharedVariableController.h"
#include "criticalSectionLock.h"
#include <assert.h>
#include "threadData.h"
#include <windows.h>
#include "traceOrReplay.h"
#include <stddef.h>
#include <string>

sharedVariableController::sharedVariableController(std::string traceFileName_)
:	writing(false), activeReaders(0), totalReaders(0), requiredReaders(0),
	version(0),	traceFileName(traceFileName_), 
	traceFileNotOpened(true) // true until traceFile is opened
{
	char* env = getenv("MODE");
	if (env!=NULL) {
		if (!(strcmp(env,"TRACE")))
			mode = TRACE;
		else if (!(strcmp(env,"REPLAY")))
			mode = REPLAY;
		else if (!(strcmp(env,"TEST")))
			mode = TEST;
		else if (!(strcmp(env,"NONE")))
			mode = NONE;
		else mode = NONE;
	}
	else mode = NONE;
	if (mode==TRACE)
		 setupTrace();
	else if (mode==REPLAY)
		 inputSynSequence();
}

sharedVariableController::~sharedVariableController() {
	if (mode==TRACE) {
		traceFile.close();
		delete myTurn;
	}
	else if (mode==REPLAY) {
		replayFile.close();
		for (int i =0; i<(maxThreadID+1); i++)
			delete versionNumber[i];
	}
}

void sharedVariableController::setupTrace() {
	myTurn = new controlConditionVariable(this);
}

void sharedVariableController::trace(event e) {
	if (traceFileNotOpened) { // don't open traceFile unless it's actually needed
		traceFileNotOpened = false;
		traceFile.open(traceFileName.c_str());
		if (!traceFile) {
			std::cout << "Error opening " << traceFileName << std::endl;
			exit(1);
		}
	}
	traceFile << e;
}

std::ostream& operator<<(std::ostream& out, const event& e) {
	out << e.threadID << ' ' << e.version << ' ' << e.totalReaders << std::endl;
	return out;
}

std::istream& operator>>(std::istream& in, event& e) {
	in >> e.threadID >> e.version >> e.totalReaders;
	return in;
}


void sharedVariableController::inputSynSequence() {
	replayFile.open(traceFileName.c_str());
	if (!replayFile) {
		std::cout << "Warning: No tracefile " << traceFileName << ". Assuming replay is off." << std::endl;
		return;
	}
	event e;
	maxThreadID = 0;
	while (replayFile >> e) {
		if (e.threadID > maxThreadID)
			maxThreadID = e.threadID;
		eventList.push_back(e);
	}
	// maxThreadID is max number of threads participating in replay
	sequences.resize(maxThreadID+1);
	std::vector<event>::iterator it;
	for (it = eventList.begin(); it != eventList.end(); it++) {
		event e = *it;
		sequences[e.threadID].push_back(e);
	}

	versionNumber = new controlConditionVariable*[maxThreads+1];
	for (int i =0; i<maxThreads+1;i++)
		versionNumber[i] = new controlConditionVariable(this);

}

int sharedVariableController::getNextReadEvent(int ID) {
	event e = sequences[ID].front();
	sequences[ID].pop_front();
	return e.version;
}

void sharedVariableController::getNextWriteEvent(int ID, int& nextVersion, int& requiredReaders) {
	event e = sequences[ID].front();
	sequences[ID].pop_front();
	requiredReaders = e.totalReaders;
	nextVersion = e.version;
}

void sharedVariableController::traceStartRead() {
	enterMonitor();
	if (writing || (!waiting.empty() &&	waiting.front() == 'W')) {
		waiting.push('R');
		myTurn->waitC();
	}
	++activeReaders;
	int ID = threadData::instance().getThreadData(GetCurrentThreadId()).threadID;
	//std::cout << "R:(" << ID << "," << version << "," << "0)" << std::endl;
	event e(ID,version,0);
	trace(e);
	if (!waiting.empty() && waiting.front() == 'R') {
		waiting.pop();
		myTurn->signalC_and_exitMonitor();
		return;
	}
	exitMonitor();
}

void sharedVariableController::traceEndRead(){
	enterMonitor();
	++totalReaders;
	--activeReaders;
	if (activeReaders==0 && !waiting.empty()) {
		assert(waiting.front() == 'W');
		waiting.pop();
		myTurn->signalC_and_exitMonitor();
		return;
	}
	exitMonitor();
}

void sharedVariableController::traceStartWrite() {
	enterMonitor();
	if (writing || activeReaders>0 || !waiting.empty()) {
		waiting.push('W');
		myTurn->waitC();
	}
	writing = true;
	int ID = threadData::instance().getThreadData(GetCurrentThreadId()).threadID;
	event e(ID,version,totalReaders);
	//std::cout << "W:(" << ID << "," << version << "," << totalReaders << ")" << std::endl;
	trace(e);
	exitMonitor();
}

void sharedVariableController::traceEndWrite(){
	enterMonitor();
	totalReaders=0;
	++version;
	writing=false;
	if (!waiting.empty()) {
		waiting.pop();
		myTurn->signalC_and_exitMonitor();
		return;
	}
	exitMonitor();
}

void sharedVariableController::requestReadPermit() {
	enterMonitor();
	int ID = threadData::instance().getThreadData(GetCurrentThreadId()).threadID;
	int nextVersion = getNextReadEvent(ID);
	if (nextVersion != version) {
		std::map<int, std::list<int> >::iterator pos;
		pos = requestingReadersMap.find(nextVersion);
		if (pos != requestingReadersMap.end()) // readers already waiting, so add to list
			pos->second.push_back(ID);
		else { // no readers waiting, make new list and add reader
			std::list<int> m; 
			m.push_back(ID);
			requestingReadersMap.insert(std::make_pair(nextVersion,m));  
		}
		versionNumber[ID]->waitC();
		pos = requestingReadersMap.find(nextVersion);
		if (pos != requestingReadersMap.end()) { // cascaded wakeup of readers (one by one)
			int nextThread = pos->second.front();
			pos->second.pop_front();
			if (pos->second.empty())
				requestingReadersMap.erase(pos);
			versionNumber[nextThread]->signalC_and_exitMonitor();
			return;
		}
	}
	exitMonitor();
}

void sharedVariableController::releaseReadPermit() {
	enterMonitor();
	++totalReaders;
	//std::cout << " in releaseReadPermit" << std::endl;
	//std::cout << " total readers= " << totalReaders << std::endl;
	//std::cout << "requiredReaders=" << requiredReaders << " totalReaders=" << totalReaders << std::endl;
	if (totalReaders==requiredReaders) { // wakeup the waiting writer
		std::map<int, int>::iterator pos;
		pos = requestingWritersMap.find(version);
		assert(pos != requestingWritersMap.end()); // must be waiting writer if requiredReaders>0
		int nextThread = pos->second;	// get ID of waiting writer
		requestingWritersMap.erase(pos);
		versionNumber[nextThread]->signalC_and_exitMonitor(); // signal waiting writer
		//std::cout << " leaving releaseReadPermit" << std::endl;
		//std::cout << "requiredReaders = " << sequences[2].front().totalReaders << " version = "  << sequences[2].front().version << std::endl;
		return;
	}
	exitMonitor();
}

void sharedVariableController::requestWritePermit() {
	enterMonitor();
	int ID = threadData::instance().getThreadData(GetCurrentThreadId()).threadID;
	int nextVersion, requiredReaders; // local requiredReaders since different for each version
	getNextWriteEvent(ID,nextVersion,requiredReaders);
	//std::cout << "nextVersion=" << nextVersion << " version=" << version << std::endl;
	if (nextVersion != version) {
		requestingWritersMap.insert(std::make_pair(nextVersion,ID)); 
		versionNumber[ID]->waitC();
	}
	//std::cout << "requiredReaders=" << requiredReaders << " totalReaders=" << totalReaders << std::endl;
	if (requiredReaders != totalReaders) {
		requestingWritersMap.insert(std::make_pair(nextVersion,ID)); 
		sharedVariableController::requiredReaders = requiredReaders; // make it global so readers can see it
		versionNumber[ID]->waitC(); // only one writer can be waiting here
	}
	exitMonitor();
}

void sharedVariableController::releaseWritePermit() {
	enterMonitor();
	totalReaders = 0;
	++version;
	requiredReaders = -1;
	std::map<int, std::list<int> >::iterator pos;
	pos = requestingReadersMap.find(version);
	if (pos != requestingReadersMap.end()) { // readers first, if any waiting for this version
		int nextThread = pos->second.front();
		pos->second.pop_front();
		if (pos->second.empty())
			requestingReadersMap.erase(pos);
		versionNumber[nextThread]->signalC_and_exitMonitor();
		return;
	}
	else { // wakeup waiting writer, if anyc (may have to wait for readers)
		std::map<int,int>::iterator pos;
		pos = requestingWritersMap.find(version);
		if (pos != requestingWritersMap.end()) {
			int nextThread = pos->second;
			requestingWritersMap.erase(pos);
			versionNumber[nextThread]->signalC_and_exitMonitor();
			return;
		}
	}
	exitMonitor();
}
