
#ifndef SEMAPHORE_H
#define SEMAPHORE_H

#include <windows.h>
#include "criticalSectionLock.h"
#include "win32Critical_Section.h"
#include "win32Semaphore.h"
#include "mutexLocker.h"
#include "semaphoreAndLockTracingAndReplay.h"

struct waitRecord {
	win32Semaphore* s;
	int ID;
};

class semaphore {
public:
	virtual void P() = 0;
	virtual void V() = 0;
	virtual void VP(semaphore* vSem) = 0;

	long getSemaphoreID() { return semaphoreID;} 
	// access method that returns  the semaphore's ID. Used in VP
	// to compare the IDs for the two semaphores, as in:
	// if (this->getSemaphoreID() > vSem->getSemaphoreID()) ...
	
	semaphore(int initialPermits) {
	  permits = initialPermits;
	  // constructor calls getID() to get an ID for the semaphore. 
	  // IDs are used in VP() to prevent circular waiting condition (i.e. deadock)
	  semaphoreID = getID();  // getId() is a static function for distributing IDs
	}
	
	virtual ~semaphore() {};
	
protected:
	// binarySemaphore and countingSemaphore inherit from
	// semaphore but they need to also be friends of semaphore 
	// so they can call lock and unlock.
	// For example:
	//  void VP(semaphore* vSem) {
	//     ...
	//     vSem->lock();  // not legal unless binarySemaphore is a friend
	//                    // since lock is protected. (Or just add a publc
	//                    // access method for permits or "willBlock()"	    
	//  }
	friend class binarySemaphore;
	friend class countingSemaphore;
	virtual bool isBinarySemaphore()=0;
	// friend class countingSemaphore  - if countingSemaphore had a VP
	
	void lock();	// to lock this semaphore in VP
	void unlock();	// to unlock this semaphore in VP
			// as in:
			//  first->lock();
			//  second->lock();
 
	int permits;
	long semaphoreID;	       // each semaphore has an integer ID

private:
	static long ID;            // semaphore IDs start with 1; see semaphore.cpp
	static criticalSectionLock idLock;   // lock for acessing ID; one lock for all semaphores
	
	static int getID() { // called by constructor when semaphore is created
	  long temp;
	  mutexLocker<criticalSectionLock> mutex(idLock);  
	  temp = ID++;             // increment in critical section since 
	  return temp;
	}
	
	win32Critical_Section mutex;   // one lock for each seamphore object
protected:
	static semaphoreAndLockTracingAndReplay* instance(int mode, 
		int numControllers, std::string traceFileName);
	static criticalSectionLock instanceLock;
};

#endif

